UNIT WavFile;
(*==========================================================================
  WAV file reader
  --------------------------------------------------------------------------
  written by   : Andreas Hofer,  fft@aho.ch,  http://www.aho.ch
  version      : 1.0.1,  18.3.2002
  copyright    : Andreas Hofer Software (c) 2002
  status       : only one channel 16 bit format supported for write
                 only first channel can be read
============================================================================*)

{$BOOLEVAL OFF }
{$EXTENDEDSYNTAX ON }
{$IOCHECKS ON }
{$LONGSTRINGS ON }
{$OPENSTRINGS ON }
{$TYPEDADDRESS OFF }
{$VARSTRINGCHECKS ON }
{$WRITEABLECONST OFF }

INTERFACE

uses
  SysUtils;



//--- reading WAV files ---------------------------------------------------

function WAVGetLength(const filename : String) : Longint;
{ How many samples the wave signal contains. }


procedure WAVReadSamples(const filename : String;
                         startpos : Longint;
                         var samples : array of Single);
{ Get samples from the wav file. The first sample has pos 0.
  Internally it calls High(samples) to determine how many samples
  are to be read, so use the Slice procedure to make the array as
  large as the number of samples you want. }



type
  TWAVFormatRec = record
    sampleFormat  : Word;      // only '1' is supported now
    nChannels     : Word;
    sampleRate    : Cardinal;  // in samples per second
    bytesPerSec   : Cardinal;
    blockAlign    : Word;
    bitsPerSample : Word;      // resolution
  end;

function WAVGetFormat(const filename : String) : TWAVFormatRec;
{ Gets the wav header information. }



{usage example:
var
  size : Cardinal;
  sampleRate : Cardinal;
  data : array of Single;
begin
  size := WAVGetLength('test.wav');
  SetLength(data, size);
  WAVReadSamples('test.wav', 0, data);
  sampleRate := WAVGetFormat('test.wav').sampleRate;
end;
}


//--- writing WAV files ---------------------------------------------------


procedure WriteWAVFile(const filename : String;
                       sampleRate : Cardinal;
                       const samples : array of Single);
{ - Format : 16 Bit, 1 channel
  - Scales the samples such that the highest value just fits into 16 bits }


{usage example:
var
  data : array[0..9999] of Single;
begin
  for i := 0 to 9999 do data[i] := sin(0.3*i);
  WriteWAVFile('sine.wav', 11025, data);
end;
}

IMPLEMENTATION (*************************************************************)


procedure Error;
begin
  raise EInOutError.Create('File format invalid or not supported.');
end;



type
  TChunkID = array[0..3] of Char;

  TChunkHeader = record
    id   : TChunkID;
    size : Cardinal;
  end;

  TListChunkHeader = record
    tag  : TChunkID;  // 'RIFF' od 'LIST'
    size : Cardinal;
    id   : TChunkID;
  end;


function RoundUp4(x : Longint) : Longint;
begin
  Result := (x+3) and $FFFFFFFC;
end;


function RoundUp2(x : Longint) : Longint;
begin
  Result := (x+1) and $FFFFFFFE;
end;


procedure ScanChunk(var f : File;
                    const path : array of TChunkID;
                    var dataPos, dataSize : Longint;
                    var found : Boolean);

  function ScanSubChunks(pos : Cardinal; pathIndex : Integer) : Boolean;
    var
      limit : Cardinal;
      hdr : TChunkHeader;
      id2 : TChunkID;
      next : Longint;
  begin
    Result := false;
    Seek(f,pos);
    BlockRead(f, hdr, SizeOf(hdr));
    limit := pos + 8 + hdr.size;
    pos := pos + 12;
    while (Result = false) and (pos < limit) do begin
      Seek(f,pos);
      BlockRead(f, hdr, SizeOf(hdr));
      next := RoundUp2(pos + 8 + hdr.size);
      if (hdr.id = 'LIST') and (pathIndex < High(path)) then begin
        BlockRead(f, id2, SizeOf(id2));
        if id2 = path[pathIndex] then begin
          Result := ScanSubChunks(pos, pathIndex+1);
        end;
      end
      else begin
        if (hdr.id = path[pathIndex]) and (pathIndex = High(path)) then begin
          dataPos := pos+8;
          dataSize := hdr.size;
          Result := true;
        end;
      end;
      pos := next;
    end;
  end;

  var
    hdr    : TChunkHeader;
    fileID : TChunkID;
begin
  Seek(f,0);
  BlockRead(f, hdr, SizeOf(hdr));
  if hdr.id <> 'RIFF' then Error;
  BlockRead(f, fileId, SizeOf(fileId));
  if fileID <> path[0] then Error;
  found := ScanSubChunks(0, 1);
end;


function WAVGetLength(const filename : String) : Longint;
  var
    f : File;
    pos, size : Longint;
    found : Boolean;
    fmt : TWAVFormatRec;
begin
  AssignFile(f,filename);
  Reset(f,1);
  try
    ScanChunk(f, ['WAVE', 'fmt '], pos, size, found);
    if (not found) then Error;
    if (size < SizeOf(TWAVFormatRec)) then Error;
    Seek(f,pos);
    BlockRead(f, fmt, SizeOf(fmt));
    if fmt.sampleFormat <> 1 then Error;

    ScanChunk(f, ['WAVE', 'data'], pos, size, found);
    if not found then Error;
    Result := size div fmt.blockAlign;
  finally
    CloseFile(f);
  end;
end;



function WAVGetFormat(const filename : String) : TWAVFormatRec;
  var
    f : File;
    pos, size : Longint;
    found : Boolean;
begin
  AssignFile(f,filename);
  Reset(f,1);
  try
    ScanChunk(f, ['WAVE', 'fmt '], pos, size, found);
    if (not found) then Error;
    if (size < SizeOf(TWAVFormatRec)) then Error;
    Seek(f,pos);
    BlockRead(f, Result, SizeOf(Result));
  finally
    CloseFile(f);
  end;
end;



procedure WAVReadSamples(const filename : String;
                         startpos : Longint;
                         var samples : array of Single);
  var
    f : File;
    pos, size : Longint;
    found : Boolean;
    fmt : TWAVFormatRec;
    buf : array of Byte;
    bytesPerSample : Integer;
    count : Integer;
    p,n,d,i : Integer;
    samp : Integer;
begin
  AssignFile(f,filename);
  Reset(f,1);
  try
    ScanChunk(f, ['WAVE', 'fmt '], pos, size, found);
    if (not found) then Error;
    if (size < SizeOf(TWAVFormatRec)) then Error;
    Seek(f,pos);
    BlockRead(f, fmt, SizeOf(fmt));
    if fmt.sampleFormat <> 1 then Error;

    ScanChunk(f, ['WAVE', 'data'], pos, size, found);
    if not found then Error;
    bytesPerSample := (fmt.bitsPerSample + 7) div 8;
    if (bytesPerSample <= 0) or (bytesPerSample > 2) then Error;
    count := Length(samples);
    n := size div fmt.blockAlign;
    if count > n then count := n;

    d := 65536 div fmt.blockAlign;
    if d > count then d := count;
    SetLength(buf, d*fmt.blockAlign);
    p := 0;
    while true do begin
      n := count-p;
      if n <= 0 then break;
      if n > d then n := d;

      Seek(f, pos + p*fmt.blockAlign);
      BlockRead(f, buf[0], n*fmt.blockAlign);

      case bytesPerSample of
        1 : begin
              for i := 0 to n-1 do begin
                samples[p+i] := (buf[i*fmt.blockAlign] - 128) / 128;
              end;
            end;
        2 : begin
              for i := 0 to n-1 do begin
                samp := SmallInt(Pointer(@buf[i*fmt.blockAlign])^);
                samples[p+i] := samp / 32768;
              end;
            end;
      end;

      p := p + n;
    end;

    for i := count to Length(samples)-1 do samples[i] := 0;

  finally
    CloseFile(f);
  end;
end;



procedure WriteWAVFile(const filename : String;
                       sampleRate : Cardinal;
                       const samples : array of Single);
  type
    TFileLayout = record
      riffID   : TChunkID;  // 'RIFF'
      size     : Cardinal;  // size of file minus 8
      waveID   : TChunkID;  // 'WAVE'
      fmtID    : TChunkID;  // 'fmt '
      fmtSize  : Cardinal;
      fmtData  : TWAVFormatRec;
      dataID   : TChunkID;  // 'data'
      dataSize : Cardinal;  // size of data in bytes
      // here follows the data
    end;
  var
    layout : TFileLayout;
    min,max,mean,k,c : Single;
    f : File;
    p,d,n,i : Integer;
    buf : array of SmallInt;
begin
  if Length(samples) = 0 then Exit;

  // prepare scaling of the data
  mean := 0;
  min := samples[0];
  max := samples[0];
  for i := 1 to Length(samples)-1 do begin
    mean := mean + samples[i];
    if samples[i] > max then max := samples[i];
    if samples[i] < min then min := samples[i];
  end;
  mean := mean / Length(samples);
  min := min - mean;
  max := max - mean;
  if -min > max then max := -min;
  if max = 0 then k := 0 else k := 32767 / max;
  c := -mean*k;

  // write header to file
  with layout do begin
    riffID  := 'RIFF';
    size    := SizeOf(TFileLayout) + 2*Length(samples) - 8;
    waveID  := 'WAVE';
    fmtID   := 'fmt ';
    fmtSize := sizeOf(TWAVFormatRec);
    fmtData.sampleFormat  := 1;
    fmtData.nChannels     := 1;
    fmtData.sampleRate    := sampleRate;
    fmtData.bytesPerSec   := 2*sampleRate;
    fmtData.blockAlign    := 2;
    fmtData.bitsPerSample := 16;
    dataID   := 'data';
    dataSize := 2*Length(samples);
  end;
  AssignFile(f, filename);
  Rewrite(f,1);
  try
    BlockWrite(f, layout, SizeOf(layout));

    // write sample data
    d := Length(samples);
    if d > 32768 then d := 32768;
    SetLength(buf,d);
    p := 0;
    while true do begin
      n := Length(samples) - p;
      if n <= 0 then break;
      if n > d then n := d;

      for i := 0 to n-1 do buf[i] := Round(samples[p+i]*k+c);
      BlockWrite(f, buf[0], 2*n);

      p := p + n;
    end;
  finally
    CloseFile(f);
  end;  
end;


END.
